# This file is part of Email-Reminder.
#
# Copyright (C) 2020 by Francois Marier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from EmailReminder.Event import EventModel, parseDay
from EmailReminder.Utils import readXMLValue, writeXMLValue


class WeeklyModel(EventModel):
    LABEL = 'Weekly Events'
    TYPE = 'weekly'
    COLUMN_NAMES = ["Event Name", "Day of the Week"]

    def fillNewEvent(self, event):
        super().fillNewEvent(event)
        writeXMLValue(self.document, event, 'day', 1)

    def getDataInternal(self, column, row):
        parentData = super().getDataInternal(column, row)
        if parentData:
            return parentData
        if column == 1:
            return readXMLValue(self.events[row], 'day')

        return None

    def setDataInternal(self, column, row, value):
        if column == 1:
            writeXMLValue(self.document, self.events[row], 'day',
                          parseDay(value, 7))
        else:
            super().setDataInternal(column, row, value)



# This file is part of Email-Reminder.
#
# Copyright (C) 2020 by Francois Marier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
from xml.dom.minidom import getDOMImplementation  # nosec
import xml.parsers.expat

from defusedxml import minidom

from EmailReminder.Anniversary import AnniversaryModel
from EmailReminder.Birthday import BirthdayModel
from EmailReminder.Monthly import MonthlyModel
from EmailReminder.Utils import readXMLValue, writeXMLValue
from EmailReminder.Weekly import WeeklyModel
from EmailReminder.Yearly import YearlyModel


class EventList:
    SUPPORTED_TYPES = ('birthday', 'anniversary', 'yearly', 'monthly',
                       'weekly')

    def __init__(self, simulate, verbose):
        self.stores = {}
        self.filename = None
        self.document = None
        self.user = None

        self.simulate = simulate
        self.verbose = verbose

    def _createDocument(self):
        self.document = getDOMImplementation().createDocument(
            None, 'email-reminder_user', None)

    def getUser(self):
        user = {}

        if not self.user:
            return user
        for field in ('first_name', 'last_name', 'email'):
            user[field] = readXMLValue(self.user, field)

        return user

    def setUser(self, user):
        for field in ('first_name', 'last_name', 'email'):
            if field in user:
                writeXMLValue(self.document, self.user, field, user[field])

    def load(self, filename):
        self.filename = filename
        if self.verbose:
            print(f"Loading events from {self.filename}")
        try:
            with open(self.filename, encoding='utf-8') as fh:
                try:
                    self.document = minidom.parseString(fh.read())
                except xml.parsers.expat.ExpatError as err:
                    if err.code == 3:  # no elements
                        self._createDocument()
                    else:
                        print(f"Error: '{filename}' is not a valid "
                              f"configuration file ({err})", file=sys.stderr)
                        return False
        except FileNotFoundError:
            self._createDocument()

        users = self.document.getElementsByTagName('email-reminder_user')
        self.user = users.item(0)

        events = self.user.getElementsByTagName('events')
        if events:
            events = events.item(0)
        else:
            events = self.document.createElement('events')
            self.document.documentElement.appendChild(events)

        for event in events.getElementsByTagName('event'):
            for eventType in self.SUPPORTED_TYPES:
                if event.getAttribute('type') == eventType:
                    self.model(eventType).addEvent(event)
                    break

        return True

    def save(self):
        if not self.document:
            return
        if self.simulate:
            return

        if self.verbose:
            print(f"Writing events to {self.filename}")

        with open(self.filename, mode='w', encoding='utf-8') as fh:
            fh.write(self.document.toxml())  # TODO: use toprettyxml()?

    def model(self, eventType):
        if eventType not in self.stores:
            if eventType == 'anniversary':
                self.stores[eventType] = AnniversaryModel(self.document)
            elif eventType == 'birthday':
                self.stores[eventType] = BirthdayModel(self.document)
            elif eventType == 'yearly':
                self.stores[eventType] = YearlyModel(self.document)
            elif eventType == 'monthly':
                self.stores[eventType] = MonthlyModel(self.document)
            elif eventType == 'weekly':
                self.stores[eventType] = WeeklyModel(self.document)
        return self.stores[eventType]



# This file is part of Email-Reminder.
#
# Copyright (C) 2020 by Francois Marier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from EmailReminder.Utils import readXMLValue, writeXMLValue, EMAIL_REGEX
from EmailReminder.Birthday import BirthdayModel


class AnniversaryModel(BirthdayModel):
    LABEL = 'Anniversaries'
    TYPE = 'anniversary'
    COLUMN_NAMES = ["Person 1", "Date", "Email 1", "Person 2", "Email 2"]

    def fillNewEvent(self, event):
        super().fillNewEvent(event)
        writeXMLValue(self.document, event, 'partner_name', '')
        writeXMLValue(self.document, event, 'partner_email', '')

    def getDataInternal(self, column, row):
        parentData = super().getDataInternal(column, row)
        if parentData:
            return parentData
        if column == 3:
            return readXMLValue(self.events[row], 'partner_name')
        if column == 4:
            return readXMLValue(self.events[row], 'partner_email')

        return None

    def setDataInternal(self, column, row, value):
        if column == 3:
            writeXMLValue(self.document, self.events[row], 'partner_name',
                          value)
        elif column == 4:
            if not EMAIL_REGEX.fullmatch(value):
                value = ''
            writeXMLValue(self.document, self.events[row], 'partner_email',
                          value)
        else:
            super().setDataInternal(column, row, value)



# This file is part of Email-Reminder.
#
# Copyright (C) 2020 by Francois Marier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from PySide2.QtCore import (
    Qt, QAbstractTableModel, QModelIndex)
from EmailReminder.Utils import readXMLValue, writeXMLValue


def parseDay(dayString, lastDay):
    try:
        day = int(dayString)
    except ValueError:
        return 1

    if day < 1:
        return 1
    if day > lastDay:
        return lastDay

    return day


class EventModel(QAbstractTableModel):
    TYPE = 'event'
    COLUMN_NAMES = ["Event Name"]

    def __init__(self, document):
        super().__init__()
        self.document = document
        self.events = []
        self.column_count = len(self.COLUMN_NAMES)

    def _createEvent(self):
        event = self.document.createElement('event')
        event.setAttribute('type', self.TYPE)
        self.fillNewEvent(event)

        user = self.document.getElementsByTagName('email-reminder_user')
        events = user.item(0).getElementsByTagName('events')
        events.item(0).appendChild(event)

        return event

    def fillNewEvent(self, event):
        writeXMLValue(self.document, event, 'name', '<New Event>')

        reminders = self.document.createElement('reminders')
        sameday = self.document.createElement('reminder')
        sameday.setAttribute('type', 'same day')
        reminders.appendChild(sameday)
        event.appendChild(reminders)

    def addEvent(self, event, parent=QModelIndex()):
        self.beginInsertRows(parent, len(self.events), len(self.events))
        self.events.append(event)
        self.endInsertRows()

    def rowCount(self, parent=QModelIndex()):
        # pylint: disable=unused-argument
        return len(self.events)

    def columnCount(self, parent=QModelIndex()):
        # pylint: disable=unused-argument
        return self.column_count

    def headerData(self, section, orientation, role):
        if role != Qt.DisplayRole:
            return None
        if orientation != Qt.Horizontal:
            return None
        return self.COLUMN_NAMES[section]

    def getDataInternal(self, column, row):
        if column == 0:
            return readXMLValue(self.events[row], 'name')

        return None

    def data(self, index, role=Qt.DisplayRole):
        column = index.column()
        row = index.row()

        if role in (Qt.DisplayRole, Qt.EditRole):
            return self.getDataInternal(column, row)

        return None

    def flags(self, index):
        # pylint: disable=unused-argument
        return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def getReminders(self, row):
        sameDay = False
        daysInAdvance = 0

        event = self.events[row]
        reminderList = event.getElementsByTagName('reminders')
        if reminderList:
            reminders = reminderList.item(0).getElementsByTagName('reminder')
            for reminder in reminders:
                typeAttribute = reminder.getAttribute('type')
                if typeAttribute == 'same day':
                    sameDay = True
                elif typeAttribute == 'days before':
                    if reminder.hasChildNodes() and reminder.firstChild:
                        daysInAdvance = int(reminder.firstChild.nodeValue)

        return {'name': readXMLValue(self.events[row], 'name'),
                'sameday': sameDay,
                'daysinadvance': daysInAdvance}

    def deleteReminder(self, parent, typeName):
        reminders = parent.getElementsByTagName('reminder')
        for reminder in reminders:
            if reminder.getAttribute('type') == typeName:
                parent.removeChild(reminder)

    def setReminder(self, parent, typeName, nodeValue=None):
        # Look for an existing element to modify.
        for reminder in parent.getElementsByTagName('reminder'):
            if reminder.getAttribute('type') == typeName:
                if nodeValue:
                    if reminder.hasChildNodes():
                        reminder.firstChild.nodeValue = nodeValue
                    else:
                        reminder.appendChild(
                            self.document.createTextNode(str(nodeValue)))
                return  # nothing else to do

        # Create a new element.
        reminder = self.document.createElement('reminder')
        reminder.setAttribute('type', typeName)
        if nodeValue:
            reminder.appendChild(self.document.createTextNode(str(nodeValue)))
        parent.appendChild(reminder)

    def setReminders(self, row, values):
        event = self.events[row]
        reminderList = event.getElementsByTagName('reminders')
        if not reminderList:
            reminders = self.document.createElement('reminders')
            event.appendChild(reminders)

        if not values['sameday']:
            self.deleteReminder(reminderList.item(0), 'same day')
        else:
            self.setReminder(reminderList.item(0), 'same day')

        if values['daysinadvance'] == 0:
            self.deleteReminder(reminderList.item(0), 'days before')
        else:
            self.setReminder(reminderList.item(0), 'days before',
                             values['daysinadvance'])

    def setDataInternal(self, column, row, value):
        if column == 0:
            writeXMLValue(self.document, self.events[row], 'name', value)

    def setData(self, index, value, role=Qt.EditRole):
        if role != Qt.EditRole:
            return False

        column = index.column()
        row = index.row()
        self.setDataInternal(column, row, value)
        self.dataChanged.emit(index, index, [role])
        return True

    def insertRows(self, row, count, parent=QModelIndex()):
        if count < 1:
            return False

        # These are not currently needed.
        if count > 1:
            return False
        if row != self.rowCount():
            return False

        self.addEvent(self._createEvent(), parent)
        return True

    def removeRows(self, row, count, parent=QModelIndex()):
        if count < 1:
            return False

        # This is neither implemented nor currently needed.
        if count > 1:
            return False

        self.beginRemoveRows(parent, row, row)
        event = self.events[row]
        del self.events[row]
        event.parentNode.removeChild(event)
        self.endRemoveRows()

        return True

    def append(self):
        return self.insertRows(self.rowCount(), 1)

    def delete(self, row):
        return self.removeRows(row, 1)
